
# First Party
from vcs_scraper.dict_remapper import remap_dict_keys

AZURE_DEVOPS_BRANCH_MAP = [
    [["name"], ["branch_id"]],
    [["name"], ["branch_name"]],
    [["commit", "commit_id"], ["latest_commit"]],
]

AZURE_DEVOPS_REPOSITORY_MAP = [
    [["name"], ["repository_name"]],
    [["project", "name"], ["project_key"]],
    [["id"], ["repository_id"]],
    [["web_url"], ["repository_url"]],
]


def map_azure_devops_repository(raw_repository):
    return remap_dict_keys(raw_repository, AZURE_DEVOPS_REPOSITORY_MAP)


def map_azure_devops_branch(raw_branch):
    return remap_dict_keys(raw_branch, AZURE_DEVOPS_BRANCH_MAP)



# Standard Library
import os
from typing import Dict, List

# Third Party
import requests
from atlassian import Bitbucket

# First Party
from vcs_scraper.model import Branch, Repository
from vcs_scraper.vcs_connectors.bitbucket_data_mapper import map_bitbucket_branch, map_bitbucket_repository
from vcs_scraper.vcs_connectors.vcs_connector import VCSConnector
from vcs_scraper.vcs_instances_parser import VCSInstance


class BitbucketConnector(VCSConnector):
    def __init__(self, scheme, host, port, access_token, proxy=None):
        self.url = f"{scheme}://{host}:{port}"
        self.access_token = access_token
        self.proxy = proxy
        self._api_client = None

    @staticmethod
    def create_client_from_vcs_instance(vcs_instance: VCSInstance):
        bitbucket_client = BitbucketConnector(
            host=vcs_instance.hostname,
            scheme=vcs_instance.scheme,
            port=vcs_instance.port,
            access_token=vcs_instance.token
        )
        return bitbucket_client

    @property
    def api_client(self):
        if not self._api_client:
            session = requests.Session()
            session.headers['Authorization'] = f"Bearer {self.access_token}"
            self._api_client = Bitbucket(
                url=self.url,
                session=session,
                proxies={"no_proxy": self.proxy}
            )
        return self._api_client

    def get_all_projects(self):
        try:
            return [project["key"] for project in self.api_client.project_list()]
        except (requests.exceptions.ConnectionError, requests.exceptions.ConnectTimeout, requests.exceptions.ProxyError,
                requests.exceptions.ReadTimeout, requests.exceptions.SSLError, requests.exceptions.HTTPError) as ex:
            raise ConnectionError(ex) from ex

    def project_exists(self, project_key: str) -> bool:
        return bool(self.api_client.project(project_key))

    def get_repos(self, project_key):
        return list(self.api_client.repo_all_list(project_key))

    def get_branches(self, project_key, repository_id):
        return list(self.api_client.get_branches(project_key, repository_id))

    @staticmethod
    def get_clone_url(clone_urls, name):
        for url in clone_urls:
            if url["name"] == name:
                return url["href"]
        return ""

    @staticmethod
    def export_repository(repository_information: Dict, branches_information: List[Dict],
                          vcs_instance_name: str) \
            -> Repository:
        """
        A method which generate a repositoryInfo object about a single bitbucket repository.

        :param vcs_instance_name: Name of the VCS instance to which the repository belongs
        :param repository_information: Bitbucket repository information as returned by the Bitbucket API.
        :param branches_information: Bitbucket branches information for a single repo as returned by the Bitbucket API.
        :return RepositoryInfo object
        """

        branches: List[Branch] = []
        for branch_information in branches_information:
            if os.getenv('SCAN_ONLY_MASTER_BRANCH', "true").lower() in "true":
                if branch_information["displayId"].lower() in ["main", "master"]:
                    branch = Branch(repository_id=repository_information["id"],
                                    **map_bitbucket_branch(branch_information))
                    branches.append(branch)
                    break
            else:
                branch = Branch(repository_id=repository_information["id"],
                                **map_bitbucket_branch(branch_information))
                branches.append(branch)
        http_clone_url = BitbucketConnector.get_clone_url(repository_information["links"]["clone"], "http")
        repository = Repository(branches=branches,
                                repository_url=http_clone_url,
                                vcs_instance_name=vcs_instance_name,
                                **map_bitbucket_repository(repository_information))
        return repository



# pylint: disable=no-name-in-module
# Standard Library
import dataclasses
import json
import os
from enum import Enum
from typing import List, Optional

# Third Party
from pydantic import BaseModel, validator
from pydantic.types import conint, constr

# First Party
from vcs_scraper.constants import AZURE_DEVOPS, BITBUCKET, GITHUB_PUBLIC


@dataclasses.dataclass
class Branch:
    repository_id: int
    branch_name: str
    branch_id: str
    latest_commit: str

    def json(self):
        json_branch = json.dumps(dataclasses.asdict(self))
        return json_branch


@dataclasses.dataclass
class Repository:

    repository_name: str
    repository_id: str
    repository_url: str
    project_key: str
    vcs_instance_name: str
    branches: List[Branch]

    def json(self):
        json_repo = json.dumps(dataclasses.asdict(self))
        return json_repo


class VCSProviders(str, Enum):
    AZURE_DEVOPS = AZURE_DEVOPS
    BITBUCKET = BITBUCKET
    GITHUB_PUBLIC = GITHUB_PUBLIC


class VCSInstance(BaseModel):

    name: constr(max_length=200)
    provider_type: VCSProviders
    hostname: constr(max_length=200)
    port: conint(gt=-0, lt=65536)
    scheme: str
    username: constr(max_length=200)
    token: constr(max_length=200)
    exceptions: Optional[List[str]] = []
    scope: Optional[List[str]] = []
    organization: Optional[str]

    @validator("scheme", pre=True)
    @classmethod
    def check_scheme(cls, value):
        allowed_schemes = ["http", "https"]
        if value not in allowed_schemes:
            raise ValueError(f"The scheme '{value}' must be one of the following {', '.join(allowed_schemes)}")
        return value

    @validator("organization", pre=True)
    @classmethod
    def check_organization(cls, value, values):
        if not value:
            if values["provider_type"] == AZURE_DEVOPS:
                raise ValueError("The organization field needs to be specified for Azure devops vcs instances")
        return value

    @validator("scope", pre=True)
    @classmethod
    def check_scope_and_exceptions(cls, value, values):
        if value and values["exceptions"]:
            raise ValueError("You cannot specify bot the scope and exceptions to the scan, only one setting"
                             " is supported.")
        return value

    @validator("username", pre=True)
    @classmethod
    def check_presence_of_username(cls, value, values):
        if not os.environ.get(value):
            raise ValueError(f"The username for VCS Instance {values['name']} could not be found in the "
                             f"environment variable {value}")
        return os.environ.get(value)

    @validator("token", pre=True)
    @classmethod
    def check_presence_of_token(cls, value, values):
        if not os.environ.get(value):
            raise ValueError(f"The access token for VCS Instance {values['name']} could not be found in the "
                             f"environment variable {value}")
        return os.environ.get(value)



# Standard Library
import logging
import os
from typing import Dict, List

# Third Party
from azure.devops.connection import Connection
from azure.devops.exceptions import AzureDevOpsServiceError
from azure.devops.released.core.core_client import CoreClient
from msrest.authentication import BasicAuthentication
from msrest.exceptions import ClientRequestError

# First Party
from vcs_scraper.model import Branch, Repository
from vcs_scraper.vcs_connectors.azure_devops_data_mapper import map_azure_devops_branch, map_azure_devops_repository
from vcs_scraper.vcs_connectors.vcs_connector import VCSConnector
from vcs_scraper.vcs_instances_parser import VCSInstance

logger = logging.getLogger(__name__)


class AzureDevopsConnector(VCSConnector):
    def __init__(self, scheme, host, port, access_token, organization, proxy=None):
        self.url = f"{scheme}://{host}:{port}/{organization}"
        self.access_token = access_token
        self.proxy = proxy
        self._api_connection = None
        self._core_api_client = None
        self._git_api_client = None

    @staticmethod
    def create_client_from_vcs_instance(vcs_instance: VCSInstance):
        azure_devops_client = AzureDevopsConnector(
            host=vcs_instance.hostname,
            scheme=vcs_instance.scheme,
            port=vcs_instance.port,
            access_token=vcs_instance.token,
            organization=vcs_instance.organization,
        )
        return azure_devops_client

    @property
    def api_client(self):
        if not self._api_connection:
            credentials = BasicAuthentication("", self.access_token)
            connection = Connection(base_url=self.url, creds=credentials)
            self._api_connection = connection

        return self._api_connection

    @property
    def core_api_client(self):
        if not self._core_api_client:
            self._core_api_client = self.api_client.clients.get_core_client()
        return self._core_api_client

    @property
    def git_api_client(self):
        if not self._git_api_client:
            self._git_api_client = self.api_client.clients.get_git_client()
        return self._git_api_client

    def get_all_projects(self):
        try:
            self._core_api_client = self.api_client.clients.get_core_client()
            all_projects = []
            call_results: CoreClient.GetProjectsResponseValue = self.core_api_client.get_projects()
            projects = call_results.value
            all_projects.extend([project.name for project in projects])
            while call_results.continuation_token:
                call_results: CoreClient.GetProjectsResponseValue = self.core_api_client.get_projects(
                    continuation_token=call_results.continuation_token)
                projects = call_results.value
                all_projects.extend([project.name for project in projects])

            return all_projects
        except ClientRequestError as ex:
            raise ConnectionError(ex) from ex

    def project_exists(self, project_key: str) -> bool:
        return bool(self.core_api_client.get_project(project_key))

    def get_repos(self, project_key):
        return list(repo.as_dict() for repo in self.git_api_client.get_repositories(project_key))

    def get_branches(self, project_key, repository_id):
        all_branches = []
        try:
            all_branches = list(self.git_api_client.get_branches(project=project_key, repository_id=repository_id))
            all_branches = [branch.as_dict() for branch in all_branches]
        except AzureDevOpsServiceError as azure_exception:
            logger.error(f"Failed to list branches for repository: {project_key}/{repository_id} --> {azure_exception}")
        return all_branches

    @staticmethod
    def get_clone_url(clone_urls, name):
        for url in clone_urls:
            if url["name"] == name:
                return url["href"]
        return ""

    @staticmethod
    def export_repository(repository_information: Dict, branches_information: List[Dict],
                          vcs_instance_name: str) \
            -> Repository:
        """
        A method which generate a repositoryInfo object about a single bitbucket repository.

        :param vcs_instance_name: Name of the VCS instance to which the repository belongs
        :param repository_information: Azure Devops repository information as returned by the Azure API.
        :param branches_information: Azure Devops branches information for a single repo as returned by the Azure API.
        :return Repository object
        """

        branches: List[Branch] = []
        for branch_information in branches_information:
            if os.getenv('SCAN_ONLY_MASTER_BRANCH', "true").lower() in "true":
                if branch_information["name"].lower() in ["main", "master"]:
                    branch = Branch(repository_id=repository_information["id"],
                                    **map_azure_devops_branch(branch_information))
                    branches.append(branch)
                    break
            else:
                branch = Branch(repository_id=repository_information["id"],
                                **map_azure_devops_branch(branch_information))
                branches.append(branch)

        repository = Repository(branches=branches,
                                vcs_instance_name=vcs_instance_name,
                                **map_azure_devops_repository(repository_information))

        return repository



# Standard Library
from os import getenv

# Third Party
from setuptools import setup


def get_required_packages():
    with open('requirements.txt') as f:
        return f.read().splitlines()


setup(
    install_requires=get_required_packages()
)



# Standard Library
import logging
from typing import List

# Third Party
import requests

# First Party
from vcs_scraper.common import create_celery_client, load_vcs_instances_into_map
from vcs_scraper.configuration import (
    RABBITMQ_DEFAULT_VHOST,
    RABBITMQ_QUEUES_PASSWORD,
    RABBITMQ_QUEUES_USERNAME,
    REQUIRED_ENV_VARS,
    RESC_RABBITMQ_SERVICE_HOST,
    VCS_INSTANCES_FILE_PATH
)
from vcs_scraper.constants import PROJECT_QUEUE, REPOSITORY_QUEUE, SECRET_SCANNER_TASK_NAME
from vcs_scraper.environment_wrapper import validate_environment
from vcs_scraper.model import Repository
from vcs_scraper.vcs_connectors.vcs_connector_factory import VCSConnectorFactory

logger = logging.getLogger(__name__)
env_variables = validate_environment(REQUIRED_ENV_VARS)
celery_client = create_celery_client("repository_collector", env_variables[RABBITMQ_QUEUES_USERNAME],
                                     env_variables[RABBITMQ_QUEUES_PASSWORD],
                                     env_variables[RESC_RABBITMQ_SERVICE_HOST],
                                     env_variables[RABBITMQ_DEFAULT_VHOST])


def extract_project_information(project_key, vcs_client, vcs_instance_name):
    """
    From a given vcs instance and a  project_key, this function returns a list containing generic
    info about of its repositories.
    """
    logger.info(f"Fetching repositories for project: '{project_key}' in vcs instance '{vcs_instance_name}'")
    project_repositories = vcs_client.get_repos(project_key)
    logger.info(f"Fetched {len(project_repositories)} repositories for project: '{project_key}'")
    project_tasks = []
    for repository in project_repositories:
        try:
            logger.info(f"Fetching branches for repository: '{project_key}/{repository['name']}'")
            repository_branches = vcs_client.get_branches(project_key=project_key, repository_id=repository["name"])
            task_parameters = vcs_client.export_repository(repository, repository_branches, vcs_instance_name)
            project_tasks.append(task_parameters)

            logger.info(f"{len(task_parameters.branches)} branch(es) for repository: "
                        f"'{project_key}/{repository['name']}' were fetched successfully")
        except requests.exceptions.HTTPError as http_exception:
            logger.error(
                f"Error while processing repository '{project_key}/{repository['name']}':"
                f" Unable to obtain the repository's branches: {http_exception}")
    return project_tasks


def send_tasks_to_celery_queue(task_name: str, queue_name: str, project_tasks: List[Repository]):
    for task in project_tasks:
        celery_client.send_task(task_name, kwargs={"repository": task.json()}, queue=queue_name)


@celery_client.task(Queue=PROJECT_QUEUE)
def collect_repositories(project_key, vcs_instance_name):
    """
    Celery worker which takes as input a project ID and collects the required information about all of its repos and
    their respective branches.
    """
    vcs_instances_map = load_vcs_instances_into_map(env_variables[VCS_INSTANCES_FILE_PATH])

    logger.info(f"Project processor received an azure project key: '{project_key}'")

    vcs_client = VCSConnectorFactory.create_client_from_vcs_instance(vcs_instances_map[vcs_instance_name])
    project_tasks = extract_project_information(project_key, vcs_client, vcs_instance_name)
    send_tasks_to_celery_queue(SECRET_SCANNER_TASK_NAME, REPOSITORY_QUEUE, project_tasks)




# First Party
from vcs_scraper.dict_remapper import remap_dict_keys

BITBUCKET_BRANCH_MAP = [
    [["id"], ["branch_id"]],
    [["displayId"], ["branch_name"]],
    [["latestCommit"], ["latest_commit"]],
]

BITBUCKET_REPOSITORY_MAP = [
    [["name"], ["repository_name"]],
    [["project", "key"], ["project_key"]],
    [["id"], ["repository_id"]],
]


def map_bitbucket_repository(raw_repository):
    return remap_dict_keys(raw_repository, BITBUCKET_REPOSITORY_MAP)


def map_bitbucket_branch(raw_branch):
    return remap_dict_keys(raw_branch, BITBUCKET_BRANCH_MAP)
